package Bits.Bitwise;

import static java.lang.Math.*;
import static math.math.math.*;

public class Bitwise {
    public static double And4Byte(double n1, double n2) {
        return AndBytes(n1, n2, 4d);
    }

    public static double And2Byte(double n1, double n2) {
        return AndBytes(n1, n2, 2d);
    }

    public static double AndByte(double n1, double n2) {
        return AndBytes(n1, n2, 1d);
    }

    public static double AndBytes(double n1, double n2, double bytes) {
        double byteVal, result, i;

        byteVal = 1d;
        result = 0d;

        if(n1 >= 0d && n1 < pow(2d, bytes*8d) && n2 >= 0d && n2 < pow(2d, bytes*8d)) {
            n1 = Truncate(n1);
            n2 = Truncate(n2);
            bytes = Truncate(bytes);

            for (i = 0d; i < bytes * 8d; i = i + 1d) {
                if (n1 % 2d == 1d && n2 % 2d == 1d) {
                    result = result + byteVal;
                }
                n1 = floor(n1 / 2d);
                n2 = floor(n2 / 2d);
                byteVal = byteVal * 2d;
            }
        }

        return result;
    }

    public static double Or4Byte(double n1, double n2) {
        return OrBytes(n1, n2, 4d);
    }

    public static double Or2Byte(double n1, double n2) {
        return OrBytes(n1, n2, 2d);
    }

    public static double OrByte(double n1, double n2) {
        return OrBytes(n1, n2, 1d);
    }

    public static double OrBytes(double n1, double n2, double bytes) {
        double byteVal, result, i;

        byteVal = 1d;
        result = 0d;

        if(n1 >= 0d && n1 < pow(2d, bytes*8d) && n2 >= 0d && n2 < pow(2d, bytes*8d)) {
            n1 = Truncate(n1);
            n2 = Truncate(n2);
            bytes = Truncate(bytes);

            for (i = 0d; i < bytes * 8d; i = i + 1d) {
                if (n1 % 2d == 1d || n2 % 2d == 1d) {
                    result = result + byteVal;
                }
                n1 = floor(n1 / 2d);
                n2 = floor(n2 / 2d);
                byteVal = byteVal * 2d;
            }
        }

        return result;
    }


    public static double Xor4Byte(double n1, double n2) {
        return XorBytes(n1, n2, 4d);
    }

    public static double Xor2Byte(double n1, double n2) {
        return XorBytes(n1, n2, 2d);
    }

    public static double XorByte(double n1, double n2) {
        return XorBytes(n1, n2, 1d);
    }

    public static double XorBytes(double n1, double n2, double bytes) {
        double byteVal, result, i;

        byteVal = 1d;
        result = 0d;

        if(n1 >= 0d && n1 < pow(2d, bytes*8d) && n2 >= 0d && n2 < pow(2d, bytes*8d)) {
            n1 = Truncate(n1);
            n2 = Truncate(n2);
            bytes = Truncate(bytes);

            for (i = 0d; i < bytes * 8d; i = i + 1d) {
                if (n1 % 2d != n2 % 2d) {
                    result = result + byteVal;
                }
                n1 = floor(n1 / 2d);
                n2 = floor(n2 / 2d);
                byteVal = byteVal * 2d;
            }
        }

        return result;
    }

    public static double Not4Byte(double b) {
        return NotBytes(b, 4d);
    }

    public static double Not2Byte(double b) {
        return NotBytes(b, 2d);
    }

    public static double NotByte(double b) {
        return NotBytes(b, 1d);
    }

    public static double NotBytes(double b, double length) {
        double result;

        result = 0d;

        if(b >= 0d && b < pow(2d, length*8d)) {
            b = Truncate(b);
            length = Truncate(length);

            result = pow(2d, length*8d) - b - 1d;
        }

        return result;
    }

    public static double ShiftLeft4Byte(double b, double amount) {
        return ShiftLeftBytes(b, amount, 4d);
    }

    public static double ShiftLeft2Byte(double b, double amount) {
        return ShiftLeftBytes(b, amount, 2d);
    }

    public static double ShiftLeftByte(double b, double amount) {
        return ShiftLeftBytes(b, amount, 1d);
    }

    public static double ShiftLeftBytes(double b, double amount, double length) {
        double result;

        result = 0d;

        if(b >= 0d && b < pow(2d, length*8d) && amount >= 0d && amount <= length*8d) {
            b = Truncate(b);
            amount = Truncate(amount);

            result = b * pow(2d, amount);
        }

        return result;
    }

    public static double ShiftRight4Byte(double b, double amount) {
        return ShiftRightBytes(b, amount, 4d);
    }

    public static double ShiftRight2Byte(double b, double amount) {
        return ShiftRightBytes(b, amount, 2d);
    }

    public static double ShiftRightByte(double b, double amount) {
        return ShiftRightBytes(b, amount, 1d);
    }

    public static double ShiftRightBytes(double b, double amount, double length) {
        double result;

        result = 0d;

        if(b >= 0d && b < pow(2d, length*8d) && amount >= 0d && amount <= length*8d) {
            b = Truncate(b);
            amount = Truncate(amount);

            result = Truncate(b / pow(2d, amount));
        }

        return result;
    }

    public static double RotateLeft4Byte(double w, double n) {
        double a;
        //return (w << n) | (w >> (32 - n));
        // Mask the upper bits first, then rotate.
        a = And4Byte(w, Not4Byte(ShiftLeft4Byte(1d, n)-1d));
        return Or4Byte(ShiftLeft4Byte(a, n), ShiftRight4Byte(w, (32d - n)));
    }

    public static double RotateRight4Bytes(double w, double n) {
        double a;
        // return (w >> d) | (w << (32 - n));
        // Mask away the upper bits first, then perform the shift.
        a = And4Byte(w, ShiftLeft4Byte(1d, n)-1d);
        return Or4Byte(ShiftRight4Byte(w, n), ShiftLeft4Byte(a, 32d - n));
    }
}
